#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# @Time    : 2022-08-20
# @Author  : wuxiangbin
# @Site    : www.0-9.ink
# @File    : identity.py
# @Software: zawn_utils
# @Function:
from io import BytesIO
from typing import List

from zawn_orm.database import Operation
from zawn_workflow.context.context import BaseContext


class WorkFlowIdentity(BaseContext):
    """ 工作流程身份 """

    def import_data(self, data: list, **kwargs) -> BaseContext:
        """ 导入身份信息列表 """

        # 整理用户和用户组的数据，写入数据库
        user_dict = {}
        new_data = []
        for row in data:
            user_id = row.pop('用户编号', None)
            user_name = row.get('用户姓名', None)

            # 无效的数据，或者已经存在的数据，则跳过
            if (not (user_id and user_name)) or (user_name in user_dict):
                continue

            # 整理映射数据和新的列表
            user_dict[user_name] = user_id
            new_data.append(row)

        for user_name, user_id in user_dict.items():
            # 用户数据
            self.operation_list.append(self.data_factory.work_user.new_operation(
                'insert',
                filter={},
                update=self.data_factory.work_user.to_db({'user_id': user_id, 'user_name': user_name, **kwargs}),
            ))

        for row in new_data:
            user_id = user_dict[row.pop('用户姓名')]
            # 用户组数据
            for group_name, value in row.items():
                if value in ['', '-']:
                    continue

                # 如果关联结果存在于用户字典，则以关联类型新增记录
                if value in user_dict:
                    self.operation_list.append(self.data_factory.work_group.new_operation(
                        'insert',
                        filter={},
                        update=self.data_factory.work_group.to_db({
                            'group_type': 'relation',
                            'group_name': group_name,
                            'user_id': user_id,
                            'target_id': user_dict[value],
                            **kwargs,
                        }),
                    ))

                # 如果关联结果不存在于用户字典，则以用户组类型新增记录
                else:
                    self.operation_list.append(self.data_factory.work_group.new_operation(
                        'insert',
                        filter={},
                        update=self.data_factory.work_group.to_db({
                            'group_type': 'group',
                            'group_name': group_name,
                            'user_id': user_id,
                            'target_id': '_',
                            **kwargs,
                        }),
                    ))

        return self

    def import_xlsx(self, xlsx_data: BytesIO, **kwargs) -> BaseContext:
        """ 导入身份信息表格 """

        # 读取xlsx文件，读取失败将返回0
        # 清空所有用户和用户组数据，重新写入最新数据

        # 读取xlsx文件，读取失败将返回0
        data = self.read_xlsx(xlsx_data, 0)
        if not data:
            return self

        # 清空所有用户和用户组数据，重新写入最新数据
        self.operation_list: List[Operation] = [
            self.data_factory.work_user.new_operation(
                'delete',
                filter=kwargs,
                update={}
            ),
            self.data_factory.work_group.new_operation(
                'delete',
                filter=kwargs,
                update={}
            ),
        ]

        self.import_data(data, **kwargs)
        return self



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# @Time    : 2022-09-01
# @Author  : wuxiangbin
# @Site    : www.0-9.ink
# @File    : state.py
# @Software: zawn_utils
# @Function:


class WorkFlowState(object):
    """ 工作流程流程 """

    STATE_SUBMIT: str = 'S1'  # 已提交状态
    STATE_DELETE: str = 'S2'  # 已删除状态

    instance = None
    state_dict = None

    def __new__(cls, *args, **kwargs):
        if cls.instance is None:
            cls.instance = super().__new__(cls, *args, **kwargs)
            cls.state_dict = cls.get_state_dict()
        return cls.instance

    @classmethod
    def get_state_dict(cls, **kwargs):
        """ 获取状态字典 """
        return {
            'A1': [  # 提交
                {'from': '_', 'to': 'S1'},
            ],
            'A2': [  # 删除
                {'from': 'S1', 'to': 'S2'},
            ],
        }



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# @Time    : 2022-08-31
# @Author  : wuxiangbin
# @Site    : www.0-9.ink
# @File    : test_searcher.py
# @Software: zawn_utils
# @Function:
import unittest

from zawn_workflow.context.searcher import WorkFlowSearcher
from zawn_workflow.test.test import TestBase


class TestSearcher(TestBase):
    """ 查询器模块测试 """

    async def test__searcher__search_user(self):
        """ 测试用户信息查询 """

        work_flow_searcher = WorkFlowSearcher()
        user_id_list = ['U0006', 'U0007']

        meta, data = await (await work_flow_searcher.search_user(user_id_list, **self.other_filter)).search()

        self.assertEqual(meta['total'], 2)
        self.assertIn(data[0]['user_name'], ['成铭衔05', '成铭衔06'])

    async def test__searcher__search_model(self):
        """ 测试工作模型查询 """

        work_flow_searcher = WorkFlowSearcher()
        model_key = 'test_workflow_01'

        meta, data = await (await work_flow_searcher.search_model(model_key=model_key, **self.other_filter)).search()

        self.assertEqual(meta['total'], 1)
        self.assertEqual(data[0]['model_key'], model_key)

    async def test__searcher__search_model_detail(self):
        """ 测试工作模型详情查询 """

        work_flow_searcher = WorkFlowSearcher()
        model_id = 'WM0001'

        meta, data = await (
            await work_flow_searcher.search_model_detail(model_id=model_id, **self.other_filter)).search()

        self.assertEqual(meta['total'], 1)
        self.assertEqual(len(data[0]['nodes']), 11)


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# @Time    : 2022-09-02
# @Author  : wuxiangbin
# @Site    : www.0-9.ink
# @File    : exception.py
# @Software: zawn_utils
# @Function:


class ModelNotFoundError(Exception):
    """ 模型未找到错误 """
    pass


class NodeNotFoundError(Exception):
    """ 节点未找到错误 """
    pass


class NodeTickError(Exception):
    """ 节点嘀嗒执行错误 """
    pass



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# @Time    : 2022-08-20
# @Author  : wuxiangbin
# @Site    : www.0-9.ink
# @File    : searcher.py
# @Software: zawn_utils
# @Function:
from typing import List

from zawn_helper.search_helper import SearchHelper
from zawn_workflow.context.context import BaseContext


class WorkFlowSearcher(BaseContext):
    """ 工作流程查询器 """

    async def search_user(self, user_id_list: List[str], **kwargs):
        """ 查询工作用户 """

        self.query = self.data_factory.work_user.new_query({
            'user_id': {'$in': user_id_list},
            **kwargs,
        })

        return self

    async def search_group(self):
        """ 查询工作用户组 """
        pass

    async def search_model(self, model_id: str = None, model_key: str = None, **kwargs):
        """ 查询工作模型 """

        filter = kwargs
        if model_id:
            filter['model_id'] = model_id
        if model_key:
            filter['model_key'] = model_key

        self.query = self.data_factory.work_model.new_query(filter)

        return self

    async def search_model_detail(self, model_id: str, **kwargs):
        """ 查询工作模型详情 """

        work_model = self.data_factory.work_model
        work_node = self.data_factory.work_node
        work_form = self.data_factory.work_form

        filter = {
            **kwargs,
            'model_id': model_id,
            '$pipeline': [
                *SearchHelper.join(
                    work_node.table_name, 'model_id', 'model_id',
                    'nodes',
                ),
                *SearchHelper.join(
                    work_form.table_name, 'nodes.form_label', 'form_label',
                    'forms',
                ),
            ],
        }
        self.query = work_model.new_query(filter)

        return self

    async def search_node(self):
        """ 查询工作节点 """
        pass

    async def search_form(self):
        """ 查询工作表单 """
        pass

    async def search_instance(self):
        """ 查询工作实例 """
        pass

    async def search_task(self):
        """ 查询工作任务 """
        pass

    async def search_value(self):
        """ 查询工作表单值 """
        pass



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# @Time    : 2022-08-20
# @Author  : wuxiangbin
# @Site    : www.0-9.ink
# @File    : repository.py
# @Software: zawn_utils
# @Function:
import datetime
from io import BytesIO
from typing import List

from zawn_orm.database import Operation
from zawn_workflow.context.base import NodeFactory
from zawn_workflow.context.context import BaseContext


class WorkFlowRepository(BaseContext):
    """ 工作流程仓库 """

    def import_form_data(self, form_data: list, **kwargs) -> BaseContext:
        """ 导入表单数据 """

        now = datetime.datetime.now().astimezone()
        op = kwargs.get('op', 'system')
        work_form = self.data_factory.work_form

        form_dict = {}
        form_label_list = []
        for row in form_data:
            form_id = row['表单编号']
            form_label = row['表单标签']
            form_dict[form_id] = {
                'form_id': form_id,
                'form_label': form_label,
                'form_name': row['表单字段'],
                'form_type': row['表单类型'],
                'placeholder': row['占位符'],
                'disabled': row['能否编辑'] in ['Y', '能', '是'],
                'required': row['是否必填'] in ['Y', '能', '是'],
                **kwargs,
            }
            form_label_list.append(form_label)

        # 更新旧表单的状态为已删除
        self.operation_list.append(work_form.new_operation(
            'update',
            filter={'form_label': {'$in': list(set(form_label_list))}, **kwargs},
            update={'$set': work_form.to_db({'status': work_form.state.DELETE, 'update_at': now, 'update_by': op})},
        ))

        # 新增表单
        for _, form in form_dict.items():
            self.operation_list.append(work_form.new_operation(
                'insert',
                filter={},
                update=work_form.to_db(form)
            ))

        return self

    def import_model_data(
            self, model_data: list, **kwargs) -> BaseContext:
        """ 导入模型列表 """

        work_model = self.data_factory.work_model
        model_dict = {}
        for row in model_data:
            model_id = row['模型编号']
            model_dict[model_id] = {
                'model_id': model_id,
                'model_key': row['模型KEY'],
                'model_type': row['模型分类'],
                'model_name': row['模型名称'],
                **kwargs,
            }

        for _, v in model_dict.items():
            self.operation_list.append(work_model.new_operation(
                'insert',
                filter={},
                update=work_model.to_db(v)
            ))

        return self

    def import_node_data(self, node_data: list, **kwargs) -> BaseContext:
        """ 导入节点列表 """

        work_node = self.data_factory.work_node
        node_dict = {}
        for i, row in enumerate(node_data):
            node_id = row['节点编号']
            model_id = row['模型编号']
            if model_id not in node_dict:
                node_dict[model_id] = []
            node_dict[model_id].append({
                'node_id': node_id,
                'parent_id': row['父节点编号'],
                'node_name': row['节点名称'],
                'node_type': row['节点类型'],
                'sort': i,
                'form_label': row['表单标签'],
                'model_id': model_id,
                **kwargs,
            })

        for model_id, node_list in node_dict.items():
            context = NodeFactory.make_node(node_list)
            if context is None:
                continue

            for node in context.root_node.to_node_list():
                self.operation_list.append(work_node.new_operation(
                    'insert',
                    filter={},
                    update=work_node.to_db(node)
                ))

        return self

    def import_xlsx(self, xlsx_data: BytesIO, **kwargs):
        """ 导入定义表格 """

        form_data = self.read_xlsx(xlsx_data, 0)
        model_data = self.read_xlsx(xlsx_data, 1)
        node_data = self.read_xlsx(xlsx_data, 2)
        if not (model_data and node_data and form_data):
            return self

        # 清空所有用户和用户组数据，重新写入最新数据
        self.operation_list: List[Operation] = [
            self.data_factory.work_model.new_operation(
                'delete',
                filter=kwargs,
                update={}
            ),
            self.data_factory.work_node.new_operation(
                'delete',
                filter=kwargs,
                update={}
            ),
            self.data_factory.work_form.new_operation(
                'delete',
                filter=kwargs,
                update={}
            ),
        ]

        self.import_form_data(form_data, **kwargs)
        self.import_model_data(model_data, **kwargs)
        self.import_node_data(node_data, **kwargs)
        return self



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# @Time    : 2022-08-21
# @Author  : wuxiangbin
# @Site    : www.0-9.ink
# @File    : test_identity.py
# @Software: zawn_utils
# @Function:
import json
import unittest
from io import BytesIO

from zawn_workflow.context.identity import WorkFlowIdentity
from zawn_workflow.data.factory import DataFactory
from zawn_workflow.test.test import TestBase

with open('../../uri.json', 'rb') as fp:
    config = json.load(fp)


class TestIdentity(TestBase):
    """ 身份模块测试 """

    async def test__identity__import_xlsx(self):
        """ 测试身份xlsx导入 """

        work_flow_identity = WorkFlowIdentity()
        with open('测试用户身份.xlsx', 'rb') as fp:
            result = await work_flow_identity.import_xlsx(
                xlsx_data=BytesIO(fp.read()),
                **self.other_filter,
            ).execute()

        data_factory = DataFactory()
        self.assertIn(data_factory.work_user.table_name, result)
        self.assertIn(data_factory.work_group.table_name, result)


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# @Time    : 2022-08-01
# @Author  : wuxiangbin
# @Site    : www.0-9.ink
# @File    : base.py
# @Software: property
# @Function:
import logging
from abc import ABC
from typing import Union, List, Optional, Type

from zawn_workflow.data.exception import NodeNotFoundError


class NodeType(object):
    """ 节点类型 """

    root = 'root'  # 根
    originate = 'originate'  # 发起
    approval = 'approval'  # 审批
    serial = 'serial'  # 串行
    join = 'join'  # 会签
    parallel = 'parallel'  # 并行
    condition = 'condition'  # 条件
    loop = 'loop'  # 循环


class NodeStatus(object):
    """ 节点状态 """

    FUTURE: str = 'FUTURE'  # 现在还没有走到的节点状态
    WAITING: str = 'WAITING'  # 只有复杂节点有该状态，表示在等待子节点审批
    READY: str = 'READY'  # 可以进行审批操作的简单节点是Ready状态
    SKIP: str = 'SKIP'  # 当一个并行节点的子节点状态为非(Ready, Waiting)时，其它兄弟节点及其子节点的状态被置为Skip
    COMPLETE: str = 'COMPLETE'  # 已经审批完成的节点状态


class HandlerType(object):
    """ 处理人类型 """

    STATIC: str = 'STATIC'  # 设计阶段静态录入人员
    FORM: str = 'FORM'  # 运行阶段表单动态录入人员
    RELATION: str = 'RELATION'  # 通过映射函数获取人员


class Context(object):
    """ 上下文类 """

    def __init__(self):
        self.root_node: Optional[Node] = None
        self._data: List[dict] = []
        self._node_dict: dict = {}

    @property
    def data(self) -> dict:
        return_data: dict = self._data.pop(0) if len(self._data) > 0 else {}
        return return_data

    @data.setter
    def data(self, value: dict):
        self._data.append(value)

    def register(self, node: 'Node'):
        """ 注册节点信息 """
        if self.root_node is None:
            self.root_node = node
        self._node_dict[node.node_id] = node

    def get_node(self, node_id: str) -> 'Node':
        """ 通过node_id获取节点 """
        return self._node_dict.get(node_id)

    def change_status(self, node_id: str, status: str):
        """ 修改节点状态 """
        try:
            node = self._node_dict[node_id]
            node.status = status
        except Exception as e:
            message = f'{e}'
            logging.error(message, exc_info=True)

    def tick(self, form_data: dict) -> 'Context':
        """ 嘀嗒方法，每帧刷新新数据 """
        self._data = form_data
        self.root_node.tick()
        return self

    def to_node_list(self, node_list: List[dict] = None, parent_id: str = '', mapping: dict = None) -> List[dict]:
        """ 转成节点列表 """
        self.root_node.to_node_list(node_list=node_list, parent_id=parent_id, mapping=mapping)


class Node(ABC):
    """ 节点类 """
    node_type = 'base'

    def __init__(self, node_id: str, node_name: str, raw_data: dict):
        self.node_id: str = node_id
        self.node_name: str = node_name
        self.parent: Optional[Node] = None
        self.children: List[Node] = []
        self.max_children: int = -1
        self._context: Optional[Context] = None
        self._status: str = NodeStatus.FUTURE
        # 处理类型、处理人id
        self._handler_type: str = HandlerType.STATIC
        self._handler_id: str = ''  # 静态时为指定id，动态时为上下文的字段名，映射时为函数名称
        # 分支条件，下标和子节点对应，不足时不执行
        self._conditions: List[str] = []
        # 存放原始数据
        self.raw_data = raw_data

    @property
    def context(self):
        return self._context

    @context.setter
    def context(self, value: Context):
        self._context = value

    @property
    def status(self):
        return self._status

    @status.setter
    def status(self, value: str):
        self._status = value

    @property
    def handler_type(self):
        return self._handler_type

    @handler_type.setter
    def handler_type(self, value: str):
        self._handler_type = value

    @property
    def handler_id(self):
        return self._handler_id

    @handler_id.setter
    def handler_id(self, value: str):
        self._handler_id = value

    @property
    def conditions(self):
        return self._conditions

    @conditions.setter
    def conditions(self, value: str):
        self._conditions = value

    def __repr__(self):
        return f'<{self.node_type} {self.node_id}.{self.node_name}>'

    def __str__(self):
        return f'{self.node_id}.{self.node_name}'

    def __eq__(self, other):
        return isinstance(other, Node) and self.node_id == other.node_id

    def set_parent(self, node: 'Node'):
        """ 设置父节点 """
        self.parent = node

    def get_children_count(self) -> int:
        """ 获取子节点数量 """
        return len(self.children)

    def is_index_valid(self, index: int) -> bool:
        """ 校验下标合法 """
        return 0 <= index < self.get_children_count()

    def get_child(self, index: int) -> Union['Node', None]:
        """ 获取子节点 """
        if index >= self.get_children_count() or index < 0:
            return None
        return self.children[index]

    def add_child(self, node: 'Node') -> 'Node':
        """ 增加子节点 """
        if 0 <= self.max_children <= self.get_children_count():
            return self
        self.children.append(node)
        return self

    def remove_child(self, node: 'Node') -> 'Node':
        """ 删除子节点 """
        for i in range(self.get_children_count()):
            if node.node_id == self.get_child(i).node_id:
                self.children.pop(i)
                break
        return self

    def to_node_list(self, node_list: List[dict] = None, parent_id: str = '', mapping: dict = None) -> List[dict]:
        """ 转成节点列表 """

        node_list = node_list or []
        mapping = mapping or {}

        id_field_name = mapping.get('node_id') or 'node_id'
        name_field_name = mapping.get('node_name') or 'node_name'
        type_field_name = mapping.get('node_type') or 'node_type'
        sort_field_name = mapping.get('sort') or 'sort'
        parent_field_name = mapping.get('parent_id') or 'parent_id'
        status_field_name = mapping.get('status') or 'status'
        handler_type_field_name = mapping.get('handler_type') or 'handler_type'
        handler_id_field_name = mapping.get('handler_id') or 'handler_id'
        conditions_field_name = mapping.get('conditions') or 'conditions'

        data = self.raw_data
        data.update({
            id_field_name: self.node_id,
            name_field_name: self.node_name,
            type_field_name: self.node_type,
            sort_field_name: len(node_list) + 1,
            parent_field_name: parent_id,
            status_field_name: self.status,
            handler_type_field_name: self.handler_type,
            handler_id_field_name: self.handler_id,
            conditions_field_name: self.conditions,
        })
        node_list.append(data)

        for child in self.children:
            child.to_node_list(node_list=node_list, parent_id=self.node_id, mapping=mapping)

        return node_list

    def tick(self) -> str:
        """ 嘀嗒方法，每帧刷新新数据 """
        running_status = self.on_tick(self.context)
        return running_status

    def on_tick(self, context: Context) -> str:
        """ 嘀嗒执行方法 """
        return self.status

    def pre_condition(self):
        """ 前置条件，满足前置条件才能进入该节点 """
        pass

    def post_condition(self):
        """ 后置条件，满足后置条件该节点才能审批完成 """
        pass

    def pre_script(self):
        """ 前置脚本，开始审批该节点时执行 """
        pass

    def post_script(self):
        """ 后置脚本，审批完成该节点时执行 """


class RootNode(Node):
    """ 根节点 """
    node_type = NodeType.root

    def on_tick(self, context: Context) -> str:
        """ 嘀嗒执行方法 """
        # 逐一遍历每个子节点，当子节点处于完成，才会执行下一个子节点

        # 如果节点状态是已完成，则直接返回已完成
        if self.status == NodeStatus.COMPLETE:
            return self.status

        # 等待状态
        self.status = NodeStatus.WAITING

        # 遍历每个子节点，执行嘀嗒方法
        # 如果嘀嗒方法返回不是完成状态，则直接返回等待状态
        for child in self.children:
            child_status = child.tick()
            if child_status != NodeStatus.COMPLETE:
                return self.status

        # 当所有子节点都是完成状态时，返回完成状态
        self.status = NodeStatus.COMPLETE
        return self.status


class OriginateNode(Node):
    """ 发起节点 """
    node_type = NodeType.originate

    def __init__(self, node_id: str, node_name: str, raw_data: dict):
        super().__init__(node_id, node_name, raw_data)
        self.max_children = 0  # 不允许添加子节点

    def on_tick(self, context: Context) -> str:
        """ 嘀嗒执行方法 """
        # 首先进入准备状态，等待审核事件后进入完成状态

        # 如果节点状态是已完成，则直接返回已完成
        if self.status == NodeStatus.COMPLETE:
            return self.status

        # 准备状态，没有发起人数据时将一直处于这个状态
        self.status = NodeStatus.READY
        handler_id = context.data.get('handler_id')
        if not handler_id:
            return self.status

        # 设置发起人信息，并标记完成状态
        self.handler_type = HandlerType.STATIC
        self.handler_id = handler_id
        self.status = NodeStatus.COMPLETE
        return self.status


class ApprovalNode(Node):
    """ 审批节点 """
    node_type = NodeType.approval

    def __init__(self, node_id: str, node_name: str, raw_data: dict):
        super().__init__(node_id, node_name, raw_data)
        self.max_children = 0  # 不允许添加子节点

    def on_tick(self, context: Context) -> str:
        """ 嘀嗒执行方法 """
        # 首先进入准备状态，等待审核事件后进入完成状态

        # 如果节点状态是已完成，则直接返回已完成
        if self.status == NodeStatus.COMPLETE:
            return self.status

        # 转换处理人信息
        if self.handler_type == HandlerType.FORM:
            self.handler_id = context.data.get(self.handler_id) or self.handler_id

        elif self.handler_type == HandlerType.RELATION:
            originator_id = context.data.get('originator_id') or ''
            command = f'{self.handler_id}("{originator_id}")'
            try:
                handler_id = eval(command, context.data)
                self.handler_id = handler_id or self.handler_id
            except Exception as e:
                logging.info(f'审批节点获取映射处理人出错:{command}:{e}')
        else:
            pass

        # 准备状态，没有发起人数据时将一直处于这个状态
        self.status = NodeStatus.READY
        return self.status


class SerialNode(RootNode):
    """ 串行节点 """
    node_type = NodeType.serial


class JoinNode(Node):
    """ 会签节点 """
    node_type = NodeType.join

    def on_tick(self, context: Context) -> str:
        """ 嘀嗒执行方法 """
        # 首先进入等待状态，等待所有子节点都是完成状态时进入完成状态

        # 进入等待状态
        self.status = NodeStatus.WAITING

        # 同时执行子节点，并检查完成状态
        status_list = []
        for child in self.children:
            child_status = child.tick()
            status_list.append(child_status == NodeStatus.COMPLETE)

        # 所有子节点状态都是完成状态时，返回完成状态
        if all(status_list):
            self.status = NodeStatus.COMPLETE
        return self.status


class ParallelNode(Node):
    """ 并行节点 """
    node_type = NodeType.parallel

    def on_tick(self, context: Context) -> str:
        """ 嘀嗒执行方法 """
        # 首先进入等待状态，等待任一子节点是完成状态时进入完成状态

        # 进入等待状态
        self.status = NodeStatus.WAITING

        # 同时执行子节点，并检查完成状态
        status_list = []
        for child in self.children:
            child_status = child.tick()
            status_list.append(child_status == NodeStatus.COMPLETE)

        # 任一子节点状态是完成状态时，返回完成状态
        if any(status_list):
            self.status = NodeStatus.COMPLETE

        # 如果是完成状态，需要把未完成的子节点设置为跳过状态
        if self.status == NodeStatus.COMPLETE:
            node_list = [self.children[i] for i in range(len(status_list)) if not status_list[i]]
            for node in node_list:
                node.status = NodeStatus.SKIP

        return self.status


class ConditionNode(Node):
    """ 条件节点 """
    node_type = NodeType.condition

    def on_tick(self, context: Context) -> str:
        """ 嘀嗒执行方法 """
        # 首先进入等待状态，等待任一子节点是完成状态时进入完成状态

        # 进入等待状态
        self.status = NodeStatus.WAITING

        # 检查条件长度，如果为0直接完成状态
        condition_length = len(self.conditions)
        if condition_length == 0:
            self.status = NodeStatus.COMPLETE
            return self.status

        # 逐一遍历每个子节点，判断条件指定情况和子节点执行情况
        status_list = [False] * len(self.children)
        for index, child in enumerate(self.children):
            # 没有判断条件的跳过
            if index >= condition_length:
                continue

            # 条件不是字符串，或空字符串的跳过
            condition = self.conditions[index]
            if not isinstance(condition, str) or condition == '':
                continue

            # 条件判断结果为false时跳过
            try:
                condition_status = bool(eval(condition, self.context.data))
                if not condition_status:
                    continue
            except Exception as e:
                logging.info(f'条件节点判断条件时出错:{index}:{condition}:{e}')

            # 执行子节点，检查执行结果是否为完成
            child_status = child.tick()
            status_list[index] = (child_status == NodeStatus.COMPLETE)

        # 任一子节点状态是完成状态时，返回完成状态
        if any(status_list):
            self.status = NodeStatus.COMPLETE

        # 如果是完成状态，需要把未完成的子节点设置为跳过状态
        if self.status == NodeStatus.COMPLETE:
            node_list = [self.children[i] for i in range(len(status_list)) if not status_list[i]]
            for node in node_list:
                node.status = NodeStatus.SKIP

        return self.status


class LoopNode(Node):
    """ 循环节点 """
    node_type = NodeType.loop

    def __init__(self, node_id: str, node_name: str, raw_data: dict):
        super().__init__(node_id, node_name, raw_data)
        self.max_children = 1  # 只允许添加1个子节点
        self.max_loop_count = 1
        self.current_loop_count = 0


class NodeFactory(object):
    """ 节点工厂 """

    # 节点类字典
    node_class_dict = {
        i.node_type: i for i in
        (RootNode, OriginateNode, ApprovalNode, SerialNode, JoinNode, ParallelNode, ConditionNode)
    }

    @classmethod
    def make_node(cls, node_list: List[dict], mapping: dict = None) -> Context:
        """ 制造节点 """

        # 实例化上下文对象
        context = Context()
        if not node_list:
            return context

        mapping = mapping or {}
        id_field_name = mapping.get('node_id') or 'node_id'
        name_field_name = mapping.get('node_name') or 'node_name'
        type_field_name = mapping.get('node_type') or 'node_type'
        sort_field_name = mapping.get('sort') or 'sort'
        parent_field_name = mapping.get('parent_id') or 'parent_id'
        status_field_name = mapping.get('status') or 'status'

        for i in sorted(node_list, key=lambda x: x.get(sort_field_name)):
            node_id = i[id_field_name]
            node_type = i[type_field_name]
            node_class: Type[Node] = cls.node_class_dict.get(node_type)
            if node_class is None:
                raise NodeNotFoundError(f'节点数据无法获取节点类:{id}:{type}')

            # 实例化节点对象
            node = node_class(
                node_id=node_id,
                node_name=i[name_field_name],
                raw_data=i,
            )
            node.status = i.get(status_field_name, NodeStatus.FUTURE)
            node.context = context
            # 加入节点字典中，用于增加子节点
            context.register(node)
            # 如果有父节点id时，把当前节点加入到父节点中
            parent_id = i[parent_field_name]
            if parent_id:
                context.get_node(parent_id).add_child(node)

        return context



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# @Time    : 2022-08-20
# @Author  : wuxiangbin
# @Site    : www.0-9.ink
# @File    : runtime.py
# @Software: zawn_utils
# @Function:
import datetime

from zawn_workflow.context.base import NodeFactory
from zawn_workflow.context.context import BaseContext
from zawn_workflow.context.searcher import WorkFlowSearcher
from zawn_workflow.data.exception import ModelNotFoundError


class WorkFlowRuntime(BaseContext):
    """ 工作流程运行时 """

    def task_tick(self, instance_id: str, task_list: list) -> BaseContext:
        """ 任务列表载入工作流程引擎执行 """

        work_task = self.data_factory.work_task
        task_field_mapping = {
            'node_id': 'task_id',
            'node_name': 'task_name',
            'node_type': 'task_type',
        }
        context = NodeFactory.make_node(task_list, mapping=task_field_mapping).tick()
        new_task_list = context.to_node_list(mapping=task_field_mapping)
        for task in new_task_list:
            self.operation_list.append(work_task.new_operation(
                'insert', filter={}, update=work_task.to_db(task),
            ))

        return self

    def copy_node_to_task(self, node_list: list, instance_id: str, **kwargs) -> BaseContext:
        """ 复制节点数据到任务数据 """

        now = datetime.datetime.now().astimezone()
        op = kwargs.get('op', 'system')

        work_instance = self.data_factory.work_instance

        # 创建任务id，组合成映射表
        id_dict = {}
        for node in node_list:
            id_dict[node['node_id']] = work_instance.new_id()

        # 组装任务操作列表
        task_list = []
        for node in node_list:
            node_id = node['node_id']
            parent_id = node['parent_id']
            task_id = id_dict[node_id]
            task = {
                'task_id': task_id,
                'parent_id': id_dict[parent_id],
                'task_name': node['node_name'],
                'task_type': node['node_type'],
                'sort': node['sort'],
                'handler_type': node['handler_type'],
                'handler_id': node['handler_id'],
                'instance_id': instance_id,
                **kwargs,
                'status': work_instance.state.SUBMIT,
                'create_at': now,
                'create_by': op,
            }
            task_list.append(task)

        # 执行工作流程
        self.task_tick(instance_id, task_list)

        return self

    def create_instance(self, model_id: str, **kwargs) -> BaseContext:
        """ 创建工作实例 """

        # 1. 复制
        work_flow_searcher = WorkFlowSearcher()
        meta, detail_list = (await work_flow_searcher.search_model_detail(model_id, **kwargs)).search()
        if meta['total'] == 0:
            raise ModelNotFoundError(f'工作模型未找到:{model_id}')

        work_instance = self.data_factory.work_instance
        work_task = self.data_factory.work_task
        work_record = self.data_factory.work_record
        work_store = self.data_factory.work_store

        for model in detail_list:
            model_id = model['model_id']
            instance_id = work_instance.new_id()
            nodes = model.pop('nodes', None) or []
            forms = model.pop('forms', None) or []
            if not nodes:
                self.logger.info(f'模型没有配置节点信息:{model_id}')
                continue

            self.operation_list.append(work_instance.new_operation(
                'insert', filter={}, update=work_instance.to_db({
                    'instance_id': instance_id,
                    'instance_name': model['model_name'],
                    'step': 0,
                    'completion_rate': 0,
                    'model_id': model_id,
                })
            ))
            self.copy_node_to_task(nodes, instance_id)

        return self

    def start(self, model_key: str, handler_id: str, **kwargs) -> BaseContext:
        """ 启动 """

        # 启动事件过程
        # 1、检查模型和用户是否存在，如果不存在则返回提示
        # 2、创建实例
        # 3、复制节点到任务
        # 4、执行一次tick
        # 5、执行一次同意时间

        work_flow_searcher = WorkFlowSearcher()
        meta, model_list = await (await work_flow_searcher.search_model(model_key=model_key, **kwargs)).search()
        if meta['total'] != 1:
            return self

        meta, user_list = await (await work_flow_searcher.search_user([handler_id], **kwargs)).search()
        if meta['total'] == 0:
            return self

        return self

    async def agree(
            self, instance_id: str, handler_id: str, form_data: dict, **kwargs):
        """ 同意事件 """
        pass

    async def reject(
            self, instance_id: str, handler_id: str, form_data: dict, **kwargs):
        """ 拒绝事件 """
        pass

    async def assign(
            self, instance_id: str, handler_id: str, form_data: dict, **kwargs):
        """ 指派事件 """
        pass

    async def entrust(
            self, instance_id: str, handler_id: str, form_data: dict, **kwargs):
        """ 委托事件 """
        pass



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# @Time    : 2022-08-22
# @Author  : wuxiangbin
# @Site    : www.0-9.ink
# @File    : context.py
# @Software: zawn_utils
# @Function:
import logging
from io import BytesIO
from typing import List, Tuple, Union

import pandas

from zawn_orm.database import Operation, Query
from zawn_orm.model import Model
from zawn_workflow.data.factory import DataFactory

response_mapping = {
    '000': (0, '操作成功'),
    '001': (0, '查询成功'),
    '002': (2, '未找到用户信息'),
    '020': (20, '找到的记录并不唯一，无法启动工作流程'),
    '999': (-1, '未知状态'),
}


class BaseContext(object):
    """ 上下文基类 """

    model = Model()
    data_factory = DataFactory()
    logger = logging

    operation_flag: bool = False  # 操作标识，True时不允许继续操作，表示有其他操作正在执行
    operation_list: List[Operation] = []
    query: Query = None

    def response_data(
            self, key: str, data: Union[list, dict], extra: list = None) -> Tuple[int, str, Union[list, dict]]:
        """ 响应数据 """

        code, message = response_mapping.get(key) or response_mapping.get('999')
        if extra and isinstance(extra, (list, tuple)):
            message = ':'.join([message, *extra])
        return code, message, data

    def read_xlsx(self, xlsx_data: BytesIO, sheet_index: int = 0) -> List[dict]:
        """ 读取xlsx文件 """
        data = pandas.read_excel(
            xlsx_data, sheet_index, engine='openpyxl', dtype='str',
        ).fillna('').to_dict(orient='records')
        return data

    async def execute(self) -> dict:
        """ 执行数据库操作 """
        if self.operation_flag:
            return {}

        self.operation_flag = True
        result = await self.model.execute(self.operation_list)
        self.operation_flag = False

        return result

    async def search(self) -> Tuple[dict, list]:
        """ 执行数据库查询 """
        data = await self.model.search(self.query)
        return data



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# @Time    : 2022-08-21
# @Author  : wuxiangbin
# @Site    : www.0-9.ink
# @File    : factory.py
# @Software: zawn_utils
# @Function:
from zawn_orm import field, model
from zawn_orm.state import BaseState


class State(BaseState):
    SUBMIT: str = 'S1'
    DELETE: str = 'S2'


class WorkUser(model.Model):
    """ 工作用户 """

    user_id = field.StringField()
    user_name = field.StringField()

    org_id = field.StringField()  # 所属机构id
    status = field.StringField()  # 状态
    create_at = field.DatetimeField()  # 创建时间
    create_by = field.StringField()  # 创建人
    update_at = field.DatetimeField()  # 更新时间
    update_by = field.StringField()  # 更新人

    state = State()  # 有限状态机


class WorkGroup(model.Model):
    """ 工作用户组 """

    group_type = field.StringField()
    group_name = field.StringField()
    user_id = field.StringField()
    target_id = field.StringField()

    org_id = field.StringField()  # 所属机构id
    status = field.StringField()  # 状态
    create_at = field.DatetimeField()  # 创建时间
    create_by = field.StringField()  # 创建人
    update_at = field.DatetimeField()  # 更新时间
    update_by = field.StringField()  # 更新人

    state = State()  # 有限状态机


class WorkForm(model.Model):
    """ 工作表单 """

    form_id = field.StringField()  # 表单id
    form_label = field.StringField()  # 表单标签，每个独立表单都有一样的标签
    form_name = field.StringField()  # 表单字段名
    form_type = field.StringField()  # 表单类型
    placeholder = field.StringField()  # 占位符
    disabled = field.BooleanField()  # 是否禁用
    required = field.BooleanField()  # 是否必填

    org_id = field.StringField()  # 所属机构id
    status = field.StringField()  # 状态
    create_at = field.DatetimeField()  # 创建时间
    create_by = field.StringField()  # 创建人
    update_at = field.DatetimeField()  # 更新时间
    update_by = field.StringField()  # 更新人

    state = State()  # 有限状态机


class WorkModel(model.Model):
    """ 工作模型 """

    model_id = field.StringField()  # 模型id
    model_key = field.StringField()  # 模型key
    model_name = field.StringField()  # 模型名称
    model_type = field.StringField()  # 模型分类类型
    version = field.IntegerField()  # 模型版本号，模型key重复时+1

    org_id = field.StringField()  # 所属机构id
    status = field.StringField()  # 状态
    create_at = field.DatetimeField()  # 创建时间
    create_by = field.StringField()  # 创建人
    update_at = field.DatetimeField()  # 更新时间
    update_by = field.StringField()  # 更新人


class WorkNode(model.Model):
    """ 工作节点 """

    node_id = field.StringField()  # 节点id
    parent_id = field.StringField()  # 父节点id
    node_name = field.StringField()  # 节点名称
    node_type = field.StringField()  # 节点类型
    sort = field.StringField()  # 排序序号
    handler_type = field.StringField()  # 处理人类型，不指定、指定、表单、关联、用户组
    handler_id = field.StringField()  # 处理人id

    form_label = field.StringField()  # 表单标签
    model_id = field.StringField()  # 模型id

    org_id = field.StringField()  # 所属机构id
    status = field.StringField()  # 状态
    create_at = field.DatetimeField()  # 创建时间
    create_by = field.StringField()  # 创建人
    update_at = field.DatetimeField()  # 更新时间
    update_by = field.StringField()  # 更新人

    state = State()  # 有限状态机


class WorkInstance(model.Model):
    """ 工作实例 """

    instance_id = field.StringField()  # 实例id
    instance_name = field.StringField()  # 实例名称，以发起人姓名+模型名称为名
    step = field.IntegerField()  # 当前步进数
    completion_rate = field.IntegerField()  # 完成率 0-100区间

    model_id = field.StringField()  # 所属模型id

    org_id = field.StringField()  # 所属机构id
    status = field.StringField()  # 状态
    create_at = field.DatetimeField()  # 创建时间
    create_by = field.StringField()  # 创建人
    update_at = field.DatetimeField()  # 更新时间
    update_by = field.StringField()  # 更新人

    state = State()  # 有限状态机


class WorkTask(model.Model):
    """ 工作任务 """

    task_id = field.StringField()  # 任务id
    parent_id = field.StringField()  # 父任务id
    task_name = field.StringField()  # 任务名称
    task_type = field.StringField()  # 任务类型
    sort = field.StringField()  # 排序序号
    handler_type = field.StringField()  # 处理人类型，不指定、指定、表单、关联、用户组
    handler_id = field.StringField()  # 处理人id

    instance_id = field.StringField()  # 所属实例id

    org_id = field.StringField()  # 所属机构id
    status = field.StringField()  # 状态
    create_at = field.DatetimeField()  # 创建时间
    create_by = field.StringField()  # 创建人
    update_at = field.DatetimeField()  # 更新时间
    update_by = field.StringField()  # 更新人

    state = State()  # 有限状态机


class WorkRecord(model.Model):
    """ 工作记录 """

    record_id = field.StringField()  # 记录id
    handler_id = field.StringField()  # 用户id
    event = field.StringField()  # 事件，同意、驳回
    comment = field.IntegerField()  # 评论
    step = field.IntegerField()  # 步进数

    task_id = field.StringField()  # 所属任务id
    instance_id = field.StringField()  # 所属实例id

    org_id = field.StringField()  # 所属机构id
    status = field.StringField()  # 状态
    create_at = field.DatetimeField()  # 创建时间
    create_by = field.StringField()  # 创建人
    update_at = field.DatetimeField()  # 更新时间
    update_by = field.StringField()  # 更新人

    state = State()  # 有限状态机


class WorkStore(model.Model):
    """ 工作贮存 """

    store_id = field.StringField()  # 贮存id
    label = field.StringField()  # 表单展示标题
    name = field.StringField()  # 表单字段名称
    type = field.StringField()  # 表单类型 包含基本类型和自定义类型
    value = field.StringField()  # 表单值 统一序列化为字符串

    record_id = field.StringField()  # 所属记录id
    task_id = field.StringField()  # 所属任务id
    instance_id = field.StringField()  # 所属实例id

    org_id = field.StringField()  # 所属机构id
    status = field.StringField()  # 状态
    create_at = field.DatetimeField()  # 创建时间
    create_by = field.StringField()  # 创建人
    update_at = field.DatetimeField()  # 更新时间
    update_by = field.StringField()  # 更新人

    state = State()  # 有限状态机


class DataFactory(object):
    """ 数据工厂 """

    work_user: model.Model = None
    work_group: model.Model = None
    work_model: model.Model = None
    work_node: model.Model = None
    work_form: model.Model = None
    work_instance: model.Model = None
    work_task: model.Model = None
    work_value: model.Model = None
    work_record: model.Model = None

    _this = None
    _instance: 'DataFactory' = None

    def __new__(cls, *args, **kwargs):
        if cls._instance is None:
            instance = super().__new__(cls, *args, **kwargs)
            instance.work_user = WorkUser()
            instance.work_group = WorkGroup()
            instance.work_model = WorkModel()
            instance.work_node = WorkNode()
            instance.work_form = WorkForm()
            instance.work_instance = WorkInstance()
            instance.work_task = WorkTask()
            instance.work_record = WorkRecord()
            instance.work_store = WorkStore()
            cls._instance = instance
        return cls._instance


if __name__ == '__main__':
    data_factory = DataFactory()
    print(id(data_factory))
    data_factory = DataFactory()
    print(id(data_factory))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# @Time    : 2022-08-02
# @Author  : wuxiangbin
# @Site    : www.0-9.ink
# @File    : test.py
# @Software: property
# @Function:

import pandas as pd

from zawn_workflow.context.base import NodeFactory, Context

node_list = [
    {
        'node_id': 'node01',
        'node_name': '串行',
        'node_type': 'root',
        'sort': 1,
        'parent_id': '',
    },
    {
        'node_id': 'node02',
        'node_name': '发起人',
        'node_type': 'originate',
        'sort': 2,
        'parent_id': 'node01',
    },
    {
        'node_id': 'node03',
        'node_name': '条件',
        'node_type': 'condition',
        'sort': 3,
        'parent_id': 'node01',
    },
    {
        'node_id': 'node04',
        'node_name': '部门主管',
        'node_type': 'approval',
        'sort': 4,
        'parent_id': 'node03',
    },
    {
        'node_id': 'node05',
        'node_name': '并行',
        'node_type': 'parallel',
        'sort': 5,
        'parent_id': 'node03',
    },
    {
        'node_id': 'node06',
        'node_name': '副总经理',
        'node_type': 'approval',
        'sort': 6,
        'parent_id': 'node05',
    },
    {
        'node_id': 'node07',
        'node_name': '总经理',
        'node_type': 'approval',
        'sort': 7,
        'parent_id': 'node05',
    },
    {
        'node_id': 'node08',
        'node_name': '会签',
        'node_type': 'join',
        'sort': 8,
        'parent_id': 'node01',
    },
    {
        'node_id': 'node09',
        'node_name': '财务总监',
        'node_type': 'approval',
        'sort': 9,
        'parent_id': 'node08',
    },
    {
        'node_id': 'node10',
        'node_name': '董事会',
        'node_type': 'approval',
        'sort': 10,
        'parent_id': 'node08',
    },
    {
        'node_id': 'node11',
        'node_name': '出纳',
        'node_type': 'approval',
        'sort': 11,
        'parent_id': 'node01',
    },
]

import json
import unittest
from io import BytesIO

from zawn_orm.impl.mongodb import MongoDB
from zawn_orm.model import Model

with open('../../uri.json', 'rb') as fp:
    config = json.load(fp)


class TestBase(unittest.IsolatedAsyncioTestCase):
    """ 测试基类 """

    async def asyncSetUp(self) -> None:
        self.other_filter = {'org_id': '_'}
        database = MongoDB()
        database.connect(config['mongodb_uri'])
        Model().set_database(database)


if __name__ == '__main__':
    context: Context = NodeFactory.make_node(
        node_list=node_list,
    )
    root_node = context.root_node
    context.data = {'originator_id': '-'}

    if root_node is None:
        raise Exception('根节点为空')

    status = root_node.tick()
    print(status)

    node_list = root_node.to_node_list()
    for i in node_list:
        print(i)

    print(node_list)

    with open('/Users/wuxiangbin/Desktop/工作流程_身份表.xlsx', 'rb') as fp:
        xlsx_data = BytesIO(fp.read())

    data = pd.read_excel(
        xlsx_data, 0, engine='openpyxl', dtype='str'
    ).fillna('-').to_dict(orient='records')
    for i in data:
        print(i, i['组A'])



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# @Time    : 2022-08-30
# @Author  : wuxiangbin
# @Site    : www.0-9.ink
# @File    : test_repository.py
# @Software: zawn_utils
# @Function:
import unittest
from io import BytesIO

from zawn_workflow.context.repository import WorkFlowRepository
from zawn_workflow.data.factory import DataFactory
from zawn_workflow.test.test import TestBase


class TestRepository(TestBase):
    """ 仓库模块测试 """

    async def test__repository__import_xlsx(self):
        """ 测试仓库xlsx导入 """

        work_flow_repository = WorkFlowRepository()
        with open('测试工作流程仓库.xlsx', 'rb') as fp:
            result = await  work_flow_repository.import_xlsx(
                xlsx_data=BytesIO(fp.read()),
                **self.other_filter,
            ).execute()

        data_factory = DataFactory()
        self.assertIn(data_factory.work_model.table_name, result)
        self.assertIn(data_factory.work_node.table_name, result)
        self.assertIn(data_factory.work_form.table_name, result)


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# @Time    : 2022-08-11
# @Author  : wuxiangbin
# @Site    : www.0-9.ink
# @File    : setup.py
# @Software: zawn_utils
# @Function:
from setuptools import setup, find_packages  # 这个包没有的可以pip一下

setup(
    name="zawn_workflow",  # 这里是pip项目发布的名称
    version="0.0.1",  # 版本号，数值大的会优先被pip
    keywords=["zawn", "workflow"],  # 关键字
    description="zawn's private utils.",  # 描述
    long_description="zawn's private utils.",
    license="MIT Licence",  # 许可证

    url="https://gitee.com/zawn/workflow.git",  # 项目相关文件地址，一般是github项目地址即可
    author="zawn",  # 作者
    author_email="zawn@qq.com",

    packages=find_packages(),
    include_package_data=True,
    platforms="any",
    install_requires=["pandas", "openpyxl", "zawn_orm"],  # 这个项目依赖的第三方库
)
